/*

Version 1.0 Copyright © 2019 - Elusien Entertainment (www.elusien.co.uk)

Creative Commons Licence This work is licensed under a Creative Commons Attribution-ShareAlike 4.0 International License.

The following is a summary of the license. For the actual wording of the license see: https://creativecommons.org/licenses/by-sa/4.0/legalcode

Attribution — You must give appropriate credit, provide a link to the license, and indicate if changes were made. You may do so in any reasonable manner, but not in any way that suggests the licensor endorses you or your use.

ShareAlike — If you remix, transform, or build upon the material, you must distribute your contributions under the same license as the original.

No additional restrictions — You may not apply legal terms or technological measures that legally restrict others from doing anything the license permits.

Notices

You do not have to comply with the license for elements of the material in the public domain or where your use is permitted by an applicable exception or limitation.

No warranties are given. The license may not give you all of the permissions necessary for your intended use. For example, other rights such as publicity, privacy, or moral rights may limit how you use the material.

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.	

*/

// Create our own transVfx namespace

var transVfx = {};

if (!Array.prototype.includes) {	// Webkit in MLT-Shotcut does not have the "includes" Array method.
  Object.defineProperty(Array.prototype, 'includes', {
    value: function(valueToFind, fromIndex) {

      if (this == null) {
        throw new TypeError('"this" is null or not defined');
      }

      // 1. Let O be ? ToObject(this value).
      var o = Object(this);

      // 2. Let len be ? ToLength(? Get(O, "length")).
      var len = o.length >>> 0;

      // 3. If len is 0, return false.
      if (len === 0) {
        return false;
      }

      // 4. Let n be ? ToInteger(fromIndex).
      //    (If fromIndex is undefined, this step produces the value 0.)
      var n = fromIndex | 0;

      // 5. If n ≥ 0, then
      //  a. Let k be n.
      // 6. Else n < 0,
      //  a. Let k be len + n.
      //  b. If k < 0, let k be 0.
      var k = Math.max(n >= 0 ? n : len - Math.abs(n), 0);

      function sameValueZero(x, y) {
        return x === y || (typeof x === 'number' && typeof y === 'number' && isNaN(x) && isNaN(y));
      }

      // 7. Repeat, while k < len
      while (k < len) {
        // a. Let elementK be the result of ? Get(O, ! ToString(k)).
        // b. If SameValueZero(valueToFind, elementK) is true, return true.
        if (sameValueZero(o[k], valueToFind)) {
          return true;
        }
        // c. Increase k by 1. 
        k++;
      }

      // 8. Return false
      return false;
    }
  });
}


(function() {
	
	this.shotcutWebVfx = ((typeof webvfx !== 'undefined') && typeof Elusien === 'undefined'); // Are we in shotcut? (true = yes, false = no)


	this.throwError = function(msg) {	// General error message routine, writes to console, sends an alert then throws an error.
		console.log(msg);
		alert(msg);
		throw new Error(msg);
	};
	
	this.resize = function() {
		var canvas    = document.getElementsByTagName("canvas")[0];
		canvas.width  = window.innerWidth;
		canvas.height = window.innerHeight;
	};
	
// Render functions provided by Dan Dennedy for WebVfx to interface with the shaderkit.js javascript library.
	
	this.GlslTransition = function(transition) {
//console.log('DEBUG: GlslTransition entered');
		this.transition   = transition;
		this.renderer     = new ShaderKit.Renderer(document.getElementsByTagName("canvas")[0]);
		this.shader       = new ShaderKit.Shader(this.renderer, transition.glsl, transition.defaultParams);
	};
	
	this.GlslTransition.prototype.render = function (time) {	// This gets called by WebVfx each time a frame is to be rendered.
		var shader = this.shader;

		var source = webvfx.getImage('sourceImage');
			
// Set the various UNIFORM variables in the shader (first the standard ones, then those specific to this translation type.
	
		shader.updateUniform("progress"  , time);
		shader.updateUniform("from"      , source.toImageData());
		shader.updateUniform("to"        , source.toImageData());
		shader.updateUniform("resolution", [source.width, source.height]);
		shader.updateUniform("ratio"     ,  source.width /source.height);
		for (var key in this.transition.uniforms) {
			shader.updateUniform(key, webvfx.getNumberParameter(key) || this.transition.uniforms[key]);
//console.log(this.transition.uniforms[key]);
		}
		
		this.renderer.render(shader);
	};
	
														// ********************************************************************
	this.init = function() {	// This is called automatically once the filter webpage has been loaded
														// ********************************************************************
		var canvas      = document.getElementsByTagName("canvas")[0];
		var filter_name = canvas.getAttribute('id');
		var data_params = canvas.getAttribute('data-params');

		if (filter_name === null) {
			throwError('ERROR: the <canvas> tag must have an "id" attribute');
		}
		
		filter_name = filter_name.toLowerCase();
console.log('filter_name=' + filter_name);

		if (! GLTransitionNamesLC.includes(filter_name)) {
			throwError('ERROR: no such filter as "' + filter_name + '"');
		}
		
		var T  = GLTransitionsByName[filter_name];
		
console.log('data-params=' + data_params);
		if (data_params) {
			this.data_param_vals = JSON.parse(data_params);
			var keys = Object.keys(T.defaultParams);
			for (var k = 0; k < keys.length; k++) {
				if (Object.keys(this.data_param_vals).includes(keys[k])) {
					T.defaultParams[keys[k]] = this.data_param_vals[keys[k]];		
				}
			}
		}
		
//console.log(T.defaultParams);

		T.glsl = '\n#ifdef GL_ES\n precision highp float;\n #endif\n uniform sampler2D from, to;\n uniform float progress;\n uniform float ratio;\n uniform float r;\n uniform vec2 resolution;\n\n'
					 + T.glsl.replace( /getFromColor\s*\((.*)\)/g, 'texture2D(from, $1)')
									 .replace( /getToColor\s*\((.*)\)/g,   'texture2D(to  , $1)') + '\n'
//					 + 'vec4 getFromColor(vec2 uv) {\n'
//					 + '  return texture2D(from, vec2(uv.x, 1.0 - uv.y));\n'
//					 + '}\n'
//					 + 'vec4 getToColor(vec2 uv) {\n'
//					 + '  return texture2D(to, vec2(uv.x, 1.0 - uv.y));\n'
//					 + '}\n'
					 + 'void main(void)\n {\n  ' 
					 + '  vec2 UV = gl_FragCoord.xy / resolution.xy;\n ' 
					 + '  gl_FragColor = transition(UV);\n '
					 + '}';
		
		transVfx.resize();
console.log(T.glsl);		
		if (!transVfx.shotcutWebVfx) {	// Debugging by running in the Browser, rather than Shotcut
			console.log(T);
			console.log(transVfx);
		}
		
		try {
//			transVfx.transition = new transVfx.GlslTransition(T); // <=== the type of transition ONLY MOVE TO HERE FOR DEBUGGING IN A BROWSER as it runs the shader program and does not destroy it, which can cause problems.
			if (transVfx.shotcutWebVfx) {	
			transVfx.transition = new transVfx.GlslTransition(T); // <=== the type of transition 
				webvfx.renderRequested.connect(transVfx.transition, transVfx.GlslTransition.prototype.render);
        webvfx.imageTypeMap = { "sourceImage" : webvfx.SourceImageType };

				webvfx.readyRender(true);
			}
		} catch (e) {
			console.log(e);
			if (this.shotcutWebVfx) {	
				webvfx.readyRender(false);
			}
		}
	};
}).apply(transVfx);

window.addEventListener("load"  , transVfx.init  , false);
window.addEventListener("resize", transVfx.resize, false);

